{-# LANGUAGE FlexibleContexts #-}

module RoomDataHandler
  ( roomDataHandler,
  )
where

import BroadcastUserData (MonadBroadcast, broadcastUserData)
import ClassyPrelude
import Control.Monad.Except (MonadError, throwError)
import Data.Aeson (eitherDecodeStrict)
import Data.Aeson.Types (FromJSON)
import GHC.IO.Exception (ExitCode (ExitSuccess))
import Network.HTTP.Types (status200, status500)
import Network.Wai
  ( ResponseReceived,
    consumeRequestBodyStrict,
    responseLBS,
  )
import State.RoomDataState
  ( MonadRoomDataStateModify (setRoomDataState),
    MonadRoomDataStateRead,
  )
import State.RoomsState
  ( roomStateDiffInOpenRooms,
    roomStateDiffers,
  )
import System.Process
import Text.Printf (printf)
import Types.AppTypes (HasConfig (getConfig))
import Types.Config (ServerOptions (..))
import Types.RoomData (RoomData, prettyPrintOpenedRoom)
import Types.WebEnv
  ( HasWebEnv (getRequest),
    getRespond,
  )

roomDataHandler ::
  ( MonadIO m,
    HasWebEnv env,
    MonadReader env m,
    MonadError ResponseReceived m,
    MonadRoomDataStateRead m,
    MonadRoomDataStateModify m,
    MonadBroadcast m,
    HasConfig env
  ) =>
  m ResponseReceived
roomDataHandler = do
  newRoomData <- parseBodyOrBadRequest
  liftIO $ putStrLn "Got triggered from prosody"
  (openedRooms, closedRooms) <- setRoomDataState newRoomData

  mapM_ notifyRoomOpend openedRooms
  mapM_ notifyRoomClosed closedRooms

  broadcastUserData
  success

parseBodyOrBadRequest ::
  ( MonadIO m,
    HasWebEnv env,
    MonadReader env m,
    MonadError ResponseReceived m,
    FromJSON a
  ) =>
  m a
parseBodyOrBadRequest = do
  liftIO $ putStrLn "Parsing body"
  body <- getRequestBody
  case eitherDecodeStrict body of
    Left errorMessage -> do
      response <- badRequest errorMessage
      throwError response
    Right newState -> return newState

getRequestBody ::
  ( MonadIO m,
    HasWebEnv env,
    MonadReader env m
  ) =>
  m ByteString
getRequestBody = do
  request <- getRequest <$> ask
  liftIO $ toStrict <$> consumeRequestBodyStrict request

badRequest ::
  ( MonadIO m,
    HasWebEnv env,
    MonadReader env m
  ) =>
  String ->
  m ResponseReceived
badRequest errorMessage = do
  respond' <- getRespond <$> ask
  liftIO $ print errorMessage
  liftIO $
    respond' $
      responseLBS
        status500
        [("Content-Type", "text/plain")]
        (fromString ("Bad request. Reason: " <> errorMessage))

success ::
  ( MonadIO m,
    HasWebEnv env,
    MonadReader env m
  ) =>
  m ResponseReceived
success = do
  respond' <- getRespond <$> ask
  liftIO $
    respond' $
      responseLBS
        status200
        [("Content-Type", "text/plain")]
        ""

notifyRoomOpend :: (MonadIO m, HasConfig a0, MonadReader a0 m) => RoomData -> m ()
notifyRoomOpend room = do
  config' <- getConfig <$> ask
  let ServerOptions {notifyExecutable = notifyExecutable'} = config'
  let (name, user) = prettyPrintOpenedRoom room
  liftIO $ printf "Room %s opened by %s\n" name user
  let command = printf "%s open '%s' '%s'" notifyExecutable' name user
  exitCode <- liftIO $ system command
  when (exitCode /= ExitSuccess) $ liftIO $ printf "Failed to notify room %s opened by %s running command %s\n" name user command

notifyRoomClosed :: (MonadIO m, HasConfig a0, MonadReader a0 m) => RoomData -> m ()
notifyRoomClosed room = do
  config' <- getConfig <$> ask
  let ServerOptions {notifyExecutable = notifyExecutable'} = config'
  let (name, _) = prettyPrintOpenedRoom room
  liftIO $ printf "Room %s closed\n" name
  exitCode <- liftIO $ system $ printf "%s closed '%s'" notifyExecutable' name
  when (exitCode /= ExitSuccess) $ liftIO $ printf "Failed to notify room %s closed\n" name
